/*    
       This software may be used and distributed according to the terms of
       the GNU General Public License (GPL), incorporated herein by
       reference.  Drivers based on this skeleton fall under the GPL and
       must retain the authorship (implicit copyright) notice.
 
       This program is distributed in the hope that it will be useful, but
       WITHOUT ANY WARRANTY; without even the implied warranty of
       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
       General Public License for more details.
 */
package cricketdaemon.reader;

import cricketdaemon.Util;
import cricketdaemon.module.*;
import cricketdaemon.reader.parse.*;
import java.util.*;

/**
 * Object that represents a beacon.  The PortReader should package all
 * the available data into the Beacon object.  The Beacon object is 
 * immutable once it leaves the PortReader.
 * 
 * @author Allen Miu
 */
public class Beacon extends Data
{
    public byte[] rawBeacon;
    public byte[][] values;
    public int seqNum;
    //public long timeStamp;
    
    private static int globalSeqNum = 0;
    private static final int MAX_SEQNUM = 65536;

    public final static Beacon invalidInstance = new Beacon(0);


    public interface BeaconHandler 
    {
	public void processBeacon(Beacon data, BitSet fieldMask);
    }

    public void process(Object handler, BitSet fieldMask)
    {
	Util.assert_nonjava(handler instanceof BeaconHandler,
		    "Beacon::process() incorrect handler type");
	BeaconHandler h = (BeaconHandler) handler;
	h.processBeacon(this, fieldMask);
    }

    public Beacon(long curTime)
    {
	super();

	// todo: may want to replace this with a hash table
	// to conserve space.
	values = new byte[ParseMap.map.length][];
	seqNum = globalSeqNum;
	globalSeqNum++;             // ensures unique beacon sequence number
	globalSeqNum %= MAX_SEQNUM; // wrap around


	timeStamp = new Long(curTime);
	timeGenerated = timeStamp;
	hwTimeStamp = timeStamp;
    }

    public void insertValue(byte[] value, int field)
    {
	// todo: could add an arugment that is a reference to the object
	// that called this function.  That way, we can enfoce that only
	// an instanceof PortReader object can call this but nothing else.
	if(field < 0 || field >= values.length) {
	    // if value already exists, do a null operation to
	    // prevent modules from tampering with beacon data.
	    Util.warn("Beacon::insertValue() seqNum="+seqNum+
		      " values["+field+"] silently ignored");
	    return;
	}
	values[field] = value;

	if(field == ParseMap.LISTENTIME)
	    hwTimeStamp = (Long) getField(ParseMap.LISTENTIME);
	
	/*
	  if(field == ParseMap.LOCALTIME)
	  timeStamp = (Long) getField(ParseMap.LOCALTIME);
	*/
    }

    public void insertInv(Object value, int field)
    {
	FieldParser p = ParseMap.map[field];
	byte[] raw = p.unparse(value);
	insertValue(raw, field);
    }
    
    // uses the default parser defined in the ParseMap
    public Object getField(int field)
    {
	if(field >= 0 && field < ParseMap.map.length)
	    return getField(field, ParseMap.map[field]);
	else
	    return null;
    }

    // uses given parser to parse field
    public Object getField(int field, FieldParser p)
    {
	if(field >= 0 && field < values.length)
	    if(values[field] != null)
		return p.parse(values[field], 0, values[field].length);
	return null;
    }

    public byte[] getFieldRaw(int field)
    {
	return values[field];
    }

    // handy for debugging: dump out the contents of a beacon
    public String toString()
    {
	String result = "Beacon: "+seqNum+" ";

	if(values == null)
	    return result;
	for(int i = 0; i < values.length; i++) {
	    if(values[i] == null)
		result += "null ";
	    else {
		for(int j = 0; j < values[i].length; j++)
		    result += Integer.toHexString(values[i][j]);
		result += " "; 
	    }
	}
	return result;
    }

    public static final String BEACON_LABEL = "Beacon:";

    public String toString2()
    {
	String result = BEACON_LABEL+" "+seqNum+" ";

	if(values == null)
	    return result;
	for(int i = 0; i < values.length; i++) {
	    if(values[i] == null)
		result += "null ";
	    else {
		Object r = getField(i); 
		result += r;
		result += " "; 
	    }
	}
	return result;
    }

    // generates a Beacon and a mask from the contents of 
    // the form generated by toString2()
    public static PlaybackBeacon parsePlayback(StringTokenizer st, 
					       long timeStamp)
    {
	Beacon beacon = null;
	BitSet mask = null;

	try {
	    String s = st.nextToken();
	    if(!s.equals(BEACON_LABEL))
		return null;
	    beacon = new Beacon(timeStamp); // warning: static seqnum gets incr
	    beacon.seqNum = Integer.parseInt(st.nextToken());

	    mask = new BitSet();
	    st.nextToken(); // pop first entry, which is always null
	    for(int i = 1; i < ParseMap.map.length; i++) {
		s = st.nextToken();
		if(!s.equals("null")) {
		    if(s.equals("(")) {
			String temp = s+" ";
			while(!(s = st.nextToken()).equals(")"))
			    temp += s+" ";
			temp += ")";
			s = temp;
		    }

		    // FIXME: need to deal with various null values and 
		    // exceptions here.

		    // FIXME: find out how to make playback files 
		    // fields generic (i.e. attrib-value?)
		    Object o = ParseMap.map[i].stringToReturnType(s);
		    byte[] b = ParseMap.map[i].unparse(o);
		    beacon.insertValue(b, i);
		    mask.set(i);
		}
	    }
	}
	catch(NoSuchElementException e) {
	    // FIXME: should check if all "logs" in catch statements
	    // have warns in them.
	    Util.warn("Beacon::parsePlayback() invalid input");
	    return null;
	}
	catch(NumberFormatException ie) {
	    Util.warn("Beacon::parsePlayback() invalid input");
	    return null;
	}

	PlaybackBeacon result = new PlaybackBeacon();
	result.beacon = beacon;
	result.mask   = mask;
	return result;
    }
}



